#ifndef ENV_C
#define ENV_C
#define GC_REDIRECT_TO_LOCAL
#include <gc/gc.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"
#include "int_funcs.h"
#include "env.h"

/* 
 *
 * Array functions
 *
 */
static inline scm_env * scm_env_new_array(scm_env * parent, unsigned int initial_capacity)
{
	scm_env * result = (scm_env *)GC_MALLOC(sizeof(scm_env));
	result->parent = parent;
	result->items.list = (struct data_item *) GC_MALLOC(sizeof(struct data_item) * initial_capacity);
	result->capacity = initial_capacity;
	result->size = 0;
	return result;
}

static inline void scm_env_add_array(scm_env * env, char * key, scm_value * value)
{
	if (env->size == env->capacity) {
		env->capacity *= 2;
		env->items.list = (struct data_item *) GC_REALLOC(env->items.list, sizeof(struct data_item) * env->capacity);
	}
	env->items.list[env->size].key = key;
	env->items.list[env->size].value = value;
	env->size++;
}

static inline struct data_item * scm_env_get_item_array(scm_env * env, char * key)
{
	int i;
	struct data_item * item = env->items.list;
	for (i = 0; i < env->size; i++) {
		if (!strcmp(item->key, key)) return item;
		item++;
	}

	return NULL;
}

static inline void scm_env_recycle_array(scm_env * env)
{
	env->size = 0;
}

static inline char ** scm_env_get_symbols_array(scm_env * env)
{
	// unsupported feature
	static char * result[] = {(char *)NULL};
	return result;
}

/*
 *
 * Hash functions
 *
 */
static inline unsigned int hash_func(char *str) 
{ 
	unsigned int hash = 5381; 

	while (*str) { 
		hash = ((hash << 5) + hash) + *str; /* hash * 33 + c */
		str++;
	}

	return hash; 
}

static inline scm_env * scm_env_new_hash(scm_env * parent, unsigned int hash_size)
{
	scm_env * result = (scm_env *)GC_MALLOC(sizeof(scm_env));
	result->parent = parent;
	result->items.hash = (struct data_item **) GC_MALLOC(sizeof(struct data_item *) * hash_size);
	result->hash_size = hash_size;
	return result;
}

static inline void scm_env_add_hash(scm_env * env, char * key, scm_value * value)
{
	struct data_item * item = (struct data_item *) GC_MALLOC(sizeof(struct data_item));
	unsigned int ix = hash_func(key) % env->hash_size;
	item->key = key;
	item->value = value;
	item->next = env->items.hash[ix];
	env->items.hash[ix] = item;
	env->size++;
}

static inline struct data_item * scm_env_get_item_hash(scm_env * env, char * key)
{
	struct data_item * item = env->items.hash[hash_func(key) % env->hash_size];
	while (item != NULL) {
		if (!strcmp(key, item->key)) return item;
		item = item->next;
	}
	return NULL;
}


static inline char ** scm_env_get_symbols_hash(scm_env * env)
{
	int i, last = 0;
	char ** result = GC_MALLOC(sizeof(char *) * env->size);
	for (i = 0; i < env->hash_size; i++) {
		struct data_item * item = env->items.hash[i];
		while (item != NULL) {
			result[last++] = item->key;
			item = item->next;
		}
	}

	return result;
}

/*
 *
 * General functions
 *
 */
scm_env * scm_env_new(scm_env * parent, unsigned int initial_capacity)
{
	return scm_env_new_array(parent, initial_capacity);
}

scm_env * scm_env_new_with_hash(scm_env * parent, unsigned int hash_size)
{
	return scm_env_new_hash(parent, hash_size);
}

void scm_env_add(scm_env * env, char * key, scm_value * value)
{
	if (env->hash_size == 0) scm_env_add_array(env, key, value);
	else scm_env_add_hash(env, key, value);
}

scm_value * scm_env_get(scm_env * env, char * key)
{
	struct data_item * item = ((env->hash_size == 0)
			? scm_env_get_item_array(env, key)
			: scm_env_get_item_hash(env, key));
	if (item != NULL) return item->value;
	
	if (env->parent == NULL) {
		char * error;
		long int value = strtol(key, &error, 10);
		if (*error != '\0') return NULL;
		scm_value * result = scm_value_new_int(value);
		scm_env_add(env, key, result);
		return result;
	}
	return scm_env_get(env->parent, key);
}

/* returns NULL if no such variable has been found */
scm_value * scm_env_set(scm_env * env, char * key, scm_value * value)
{
	struct data_item * item = ((env->hash_size == 0)
			? scm_env_get_item_array(env, key)
			: scm_env_get_item_hash(env, key));

	if (item != NULL) {
		item->value = value;
		return value;
	}
	return scm_env_set(env->parent, key, value);
}

void scm_env_recycle(scm_env * env)
{
	scm_env_recycle_array(env);
}

/* returns set of all symbols in environment with NULL at the end */ 
char ** scm_env_get_symbols(scm_env * env)
{
	if (env->hash_size == 0) return scm_env_get_symbols_array(env);
	return scm_env_get_symbols_hash(env);
}

#endif
