#define GC_REDIRECT_TO_LOCAL
#define GC_THREADS
#include <gc/gc.h>
#include <pthread.h>

#include "mutex.h"
#include "rt.h"
#include "vm.h"
#include "vm.c"
#include "types.h"
#include "env.h"
#include "initial_env.h"
#include "eval.h"
#include "int_funcs.h"

#include <sys/time.h>
#include <stdio.h>

scm_rt * rt_new(unsigned short max_threads, scm_scheduler scheduler)
{
	scm_rt * rt = (scm_rt *) GC_MALLOC(sizeof(scm_rt));
	rt->sched_count = 0;
	rt->max_threads = max_threads;
	rt->scheduler_limit = 100000;
	//rt->scheduler_limit = 1000;
	rt->reopen_limit = 100;
	rt->env = get_initial_env(NULL);
	rt->scheduler = scheduler;
	mutex_init(rt->lock);
	return rt; 
}

scm_value * rt_eval(scm_rt * rt, scm_value * value)
{
	scm_vm * vm;
	scm_value * result;

	rt_lock(rt);
	rt->sched_count = 0;
	rt_unlock(rt);

	vm = vm_new_init(rt, vm_command_new0(OP_EVAL, rt->env, value));
	rt->main_vm = vm;
	eval_vm(vm);

	vm_lock(vm);
	if (!stack_empty(vm->rslt)) result = vm_result_pop(vm); /* result stack contains only one result which is actually the result */
	else result = scm_value_new_error("Inconsistant stack");
	vm_unlock(vm);
	return result; 
}

static void write_log_item(scm_rt * rt, FILE * f, struct thread_log * log_item, long * start)
{
	if (log_item->next == NULL) {
		*start = log_item->sec;
		return;
	}
	write_log_item(rt, f, log_item->next, start);
	if (log_item->count >= 0) {
		// FIXME: miliseconds are incorrect
		fprintf(f, "%ld.%ld\t%i\n", log_item->sec - *start, log_item->us, log_item->count + 1);
	}
	
}

static void write_log(scm_rt * rt)
{
	FILE * f = fopen("thread.log", "w+");
	long start;
	if (rt->log == NULL) return;
	write_log_item(rt, f, rt->log, &start);
	fflush(f);
	fclose(f);
}


scm_value * rt_eval_str(scm_rt * rt, char * str)
{
	scm_value * cmds = parse_input(str);
	scm_value * result = NULL;

	if (cmds == NULL) return scm_value_new_error("Parser error");
	while (SCM_TYPE(cmds) == PAIR) {
		result = rt_eval(rt, SCM_CAR(cmds)) ;
		if (SCM_TYPE(result) == ERROR) {
			fprintf(stderr, "%s\n", SCM_SYMBOL(result));
			break;
		}
		cmds = SCM_CDR(cmds);
	}
	write_log(rt);
	return result;
}

/* returns 1 if no other threads may be scheduled */
static inline int rt_schedule_on_eval(scm_rt * rt, scm_vm * vm, command * cmd)
{
	scm_vm * nvm; 
	int done = 0;
	pthread_t * thr;
	rt_lock(rt);
	if (rt->sched_count >= rt->max_threads) {
		rt_unlock(rt);
		return 1;
	} else {
		rt->sched_count++;
		rt_log(rt);
		if (rt->sched_count >= rt->max_threads) done = 1;
	}
	rt_unlock(rt);

	nvm = vm_new_init(rt, vm_command_new0(OP_EVAL, cmd->env, cmd->arg1));
	nvm->parent = vm;

	cmd->op = OP_FEVAL;
	cmd->vm = nvm;

	thr = (pthread_t *)GC_MALLOC(sizeof(pthread_t));
	pthread_create(thr, NULL, eval_vm, nvm);

	return done;
}

void backward_scheduler(scm_rt * rt, scm_vm * vm)
{
	stack_item xstack = stack_internal_get_bottom(vm->exct);
	if (stack_empty(vm->exct)) return;
	while (!stack_internal_get_done(vm->exct, xstack)) {
		command * cmd = stack_internal_get_value(vm->exct, xstack);

		if ((cmd->op == OP_EVAL) && (SCM_TYPE(cmd->arg1) == PAIR) && (cmd->tail_position != HEAD)) {
			int schedule = 1;

			if (SCM_TYPE(SCM_CAR(cmd->arg1)) == SYMBOL) {
				scm_value * test = scm_env_get(rt->env, SCM_SYMBOL(SCM_CAR(cmd->arg1)));
				//if (test && ((SCM_TYPE(test) == FUNC) || (SCM_TYPE(test) == SPEC_FORM))) {
				if (test && ((SCM_TYPE(test) == FUNC) || (SCM_TYPE(test) == SPEC_FORM) || (SCM_TYPE(test) == PRECOMPILED_MACRO))) {
					schedule = 0;
				}

			}
			if (schedule) {
				int done = rt_schedule_on_eval(rt, vm, cmd);
				if (done) break;
			}

		}
		xstack = stack_internal_get_prev(vm->exct, xstack);
	}
}

void forward_scheduler(scm_rt * rt, scm_vm * vm)
{
	//int limiter = 8;
	int limiter = 2;
	/*
	rt_lock(rt);
	if (rt->sched_count >= rt->max_threads) {
		rt_unlock(rt);
		return;
	}
	rt_unlock(rt);
	*/

	stack_item xstack = stack_internal_get_top(vm->exct);
	while (!stack_internal_get_done(vm->exct, xstack)) {
		command * cmd = stack_internal_get_value(vm->exct, xstack);

		if (cmd->op == OP_FEVAL) limiter--;
		if (cmd->op == OP_DROP) break;
		if ((cmd->op == OP_EVAL) && (SCM_TYPE(cmd->arg1) == PAIR) && (cmd->tail_position != HEAD)) {
			if (limiter == 0) {
				int schedule = 1;
				if (SCM_TYPE(SCM_CAR(cmd->arg1)) == SYMBOL) {
					scm_value * test = scm_env_get(rt->env, SCM_SYMBOL(SCM_CAR(cmd->arg1)));
					if (test && ((SCM_TYPE(test) == FUNC) || (SCM_TYPE(test) == SPEC_FORM)
						|| (SCM_TYPE(test) == PRECOMPILED_MACRO))) {
						schedule = 0;
					}

				}
				if (schedule) {
					int done = rt_schedule_on_eval(rt, vm, cmd);
					if (done) break;
				}

			} else limiter--;
		}
		xstack = stack_internal_get_next(vm->exct, xstack);
	}
}


/* Assumes locked vm, unlocked rt */
void rt_schedule_fevals(scm_rt * rt, scm_vm * vm)
{
	rt_lock(rt);
	if (rt->sched_count >= rt->max_threads) {
		rt_unlock(rt);
		return;
	}
	rt_unlock(rt);
	
	scm_scheduler scheduler = rt->scheduler;
	scheduler(rt, vm);
}


void rt_lock(scm_rt * rt)
{
	mutex_lock(rt->lock);
}

void rt_unlock(scm_rt * rt) 
{
	mutex_unlock(rt->lock);
}

void rt_log(scm_rt * rt)
{
	struct timeval tv;
	struct thread_log * log = (struct thread_log *)GC_MALLOC(sizeof(struct thread_log));
	gettimeofday(&tv, NULL);
	log->count = rt->sched_count;
	log->sec = tv.tv_sec;
	log->us = tv.tv_usec;
	log->next = rt->log;
	rt->log = log;
}
