%{
#include <stdlib.h>
#include "types.h"
#include "scheme.tab.h"

#define YYERROR_VERBOSE 1

scm_value * result = NULL;

%}

%union {
	char * symbol;
	scm_value * parsed;
}

%token <symbol> T_SYMBOL
%token <symbol> T_NUMBER
%token <symbol> T_FLOAT_NUMBER
%token T_LEFT_PAREN
%token T_RIGHT_PAREN
%token T_LEFT_BRACKET
%token T_RIGHT_BRACKET
%token T_QUOTE
%token T_UNQUOTE
%token T_UNQUOTE_SPLICING
%token T_QUASIQUOTE
%token T_DOT
%token T_VECTOR
%token T_TRUE
%token T_FALSE

%type <parsed> expressions symbol symbols list quote quasiquote unquote unquote_splicing true false

%start expressions

%%

expressions:	symbols { result = $1; }
;

symbol:	T_SYMBOL { $$ = scm_value_new_symbol($1); } 
        | T_NUMBER { $$ = scm_value_new_int(atoi($1)); } 
        | T_FLOAT_NUMBER { $$ = scm_value_new_float(atof($1)); } 
	| list
	| quote
	| quasiquote
	| unquote
	| unquote_splicing
	| true 
	| false 
;

symbols: symbol { $$ = scm_value_new_pair($1, SCM_CONST_NIL); }
	| symbol T_DOT symbol { $$ = scm_value_new_pair($1, $3); } 
	| symbol symbols { $$ = scm_value_new_pair($1, $2); }

;

list:	T_LEFT_PAREN T_RIGHT_PAREN { $$ = SCM_CONST_NIL; } 
	| T_LEFT_PAREN symbols T_RIGHT_PAREN { $$ = $2; }
	| T_LEFT_BRACKET symbols T_RIGHT_BRACKET { $$ = $2; }
	| T_VECTOR T_LEFT_PAREN T_RIGHT_PAREN { $$ = scm_value_new_vector_from_list(scm_value_new_nil()); }
	| T_VECTOR T_LEFT_PAREN symbols T_RIGHT_PAREN { $$ = scm_value_new_vector_from_list($3); }
;

false: 	T_FALSE { $$ = scm_value_new_bool(0); }
;

true: 	T_TRUE { $$ = scm_value_new_bool(1); }
;

/* Quotation symbols */
quote:	T_QUOTE symbol	{ $$ = scm_value_new_pair(
				scm_value_new_symbol("quote"), 
				scm_value_new_pair($2, SCM_CONST_NIL)); }
;

quasiquote: T_QUASIQUOTE symbol { $$ = scm_value_new_pair(
					scm_value_new_symbol("quasiquote"), 
						scm_value_new_pair($2, SCM_CONST_NIL)); }
;

unquote: T_UNQUOTE symbol { $$ = scm_value_new_pair(
				scm_value_new_symbol("unquote"), 
				scm_value_new_pair($2, SCM_CONST_NIL)); }
;

unquote_splicing: T_UNQUOTE_SPLICING symbol { $$ = scm_value_new_pair(
				scm_value_new_symbol("unquote-splicing"), 
				scm_value_new_pair($2, SCM_CONST_NIL)); }
;

%%

#include "lex.yy.c"

scm_value * parse_input(char * input)
{
	char safe_input[strlen(input) + 2]; /* ensures that input is finished with newline char */
	struct yy_buffer_state * buffer;

	strcpy(safe_input, input);
	strcat(safe_input, "\n");
	
	buffer = yy_scan_string(safe_input);
	int success = yyparse();
	yy_delete_buffer(buffer);
	return (success ? NULL : result);
}

/**
 * Checks parentheses of an expression
 * Returns (0 - Ok), (> 0 - missing right parenthesis), (< 0 - redundant right parenthesis)
 */
int check_parentheses(char * expr)
{
	char safe_input[strlen(expr) + 2]; /* ensures that input is finished with newline char */
	struct yy_buffer_state * buffer;
	int parens = 0;

	strcpy(safe_input, expr);
	strcat(safe_input, "\n");
	buffer = yy_scan_string(safe_input);
	while (1) {
		int token = yylex();
		if (token == T_LEFT_PAREN) parens++;
		if (token == T_RIGHT_PAREN) parens--;
		if (token == 0) break;
	}
	yy_delete_buffer(buffer);
	return parens;
}

int yyerror(char *s)
{
	fprintf(stderr, "%s\n", s);
	return 0;
}

int yywrap()
{
	return 1;
}
