#define GC_THREADS
#define GC_REDIRECT_TO_LOCAL
#include <gc/gc.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>

#include "types.h"
#include "int_funcs.h"
#include "rt.h"
#include "scheme.tab.h"
#include "readline_wrapper.h"
#include "eval.h"

#define BUF_SIZE	1024

#define RUN(x) (rt_eval_str(rt, x))
#define REP(x) { \
	scm_value * val = RUN(x);	\
	if (SCM_TYPE(val) != ERROR) display_value(val); \
	printf("\n"); }

#define TEST_FOR_ARG_PARAMETER {	\
	if (argc == 1) {	\
		printf("Not enough arguments.\n");	\
		return 1;	\
	}	\
}

struct arguments {
	int argc;
	char **argv;
	unsigned int load_standard_lib;
	unsigned short number_of_threads;
	int start_repl;
	char * expression;
	char * script_file;
	scm_scheduler scheduler;
};

static int process_arguments(struct arguments * args)
{
	int argc = args->argc;
	char **argv = args->argv;

	args->load_standard_lib = 1;
	args->start_repl = 1;
	args->expression = NULL;
	args->script_file = NULL;
	args->number_of_threads = 2;
	args->scheduler = backward_scheduler;

	argc--;
	argv++;
	while (argc > 0) {
		char * arg = *argv;
		if (!strcmp(arg, "-q")) args->load_standard_lib = 0;
		if (!strcmp(arg, "-s") || !strcmp(arg, "-c")) args->start_repl = 0;
		if (!strcmp(arg, "-s") || !strcmp(arg, "-l")) {
			char * file_name;
			TEST_FOR_ARG_PARAMETER;
			argc--;
			file_name = *++argv;
			args->script_file = load_file(file_name);
			if (args->script_file == NULL) {
				printf("Unable to open file: %s.\n", file_name);
				return 1;
			}
		}

		if (!strcmp(arg, "-p")) {
			TEST_FOR_ARG_PARAMETER;
			argc--;
			argv++;
			if (!strcmp(*argv, "forward")) args->scheduler = forward_scheduler;
			else if (!strcmp(*argv, "backward")) args->scheduler = backward_scheduler;
			else {
				printf("Unknown scheduler.\n");
				return 1;
			}
		}

		if (!strcmp(arg, "-e") || !strcmp(arg, "-c")) {
			TEST_FOR_ARG_PARAMETER;
			argc--;
			args->expression = *++argv;	
		}

		if (!strcmp(arg, "-t") || !strcmp(arg, "--threads")) {
			TEST_FOR_ARG_PARAMETER;
			argc--;
			args->number_of_threads = atoi(*++argv);
		}

		if (!strcmp(arg, "-h") || !strcmp(arg, "--help"))  {
			printf("Implicitly parallel scheme\n");
			printf("Usage: schemik OPTIONS\n");
			printf("  -t N, --threads N\tevalute expressions using N threads [default: 2]\n");
			printf("  -p STRATEGY\t\tscheduler strategy (forward/backward) [default: backward]\n");
			printf("  -q\t\t\tignore base library\n");
			printf("  -s FILE\t\tevaluate scheme code in FILE and exit\n");
			printf("  -l FILE\t\tevaluate scheme code in FILE and start REPL\n");
			printf("  -c EXPR\t\tevaluate expression and exit\n");
			printf("  -e EXPR\t\tevaluate expression and start REPL\n");
			printf("  -h, --help\t\tdisplay this help and exit\n");
			args->start_repl = 0;
		}
		argv++;
		argc--;
	}

	return 0;
}

void * main_thread(void * arg) {
	
	scm_rt * rt;
	int state;
	char * expr = NULL;
	struct arguments * args = (struct arguments *) arg;

	GC_INIT();
	state = process_arguments(args);
	if (state) return NULL;
	
	scm_init_constants();
	rt = rt_new(args->number_of_threads, args->scheduler);

	if (args->load_standard_lib) {
		const int READLINK_BUF_SIZE = 65535;
		char * code = NULL;
		char buf[READLINK_BUF_SIZE];
		int size = readlink("/proc/self/exe", buf, READLINK_BUF_SIZE);
		if (size) {
			*strrchr(buf, '/') = '\0';
			strcat(buf, "/scm/base.scm");
			code = load_file(buf);
		}
		if (code == NULL) code = load_file("scm/base.scm");
		if (code != NULL) RUN(code);
		else {
			fprintf(stderr, "Standard library not found\n");
		}
	}

	if (args->script_file != NULL) RUN(args->script_file);
	if (args->expression != NULL) RUN(args->expression);
	if (!args->start_repl) return NULL;

	while (1) {
		int i;
		int parentheses;

#ifdef READLINE_WRAPPER_H
		char * buf;
		char ** commands = scm_env_get_symbols(rt->env);
		if (expr == NULL) buf = readline_wrapper("schemik> ", commands);
		else buf = readline_wrapper("", commands);
		if (buf == NULL) {
			printf("\n");
			return NULL;
		}
#else
		char buf[BUF_SIZE];
		if (expr == NULL) printf("schemik> ");
		fgets(buf, BUF_SIZE, stdin);
#endif

		if (expr == NULL) expr = GC_MALLOC(strlen(buf) + 1);
		else expr = GC_REALLOC(expr, strlen(expr) + strlen(buf) + 1);
		strcat(expr, buf);

		int is_empty = 1;
		for (i = 0; i < strlen(expr); i++) {
			char c = expr[i];
			if (!(isblank(c) || iscntrl(c))) {
				is_empty = 0;
				break;
			}
		}

		if (is_empty) {
			GC_FREE(expr);
			expr = NULL;
			continue;
		}

		parentheses = check_parentheses(expr);
		if (parentheses <= 0) {
			if (parentheses == 0) REP(expr)
			else printf("Unexpected ')'.\n");
			GC_FREE(expr);
			expr = NULL;
		}
	}
	return NULL;
}

int main(int argc, char **argv) 
{
	pthread_t thr;
	struct arguments args;
	args.argc = argc;
	args.argv = argv;
	pthread_create(&thr, NULL, main_thread, &args);
	pthread_join(thr, NULL);
	return 0;
}
