#define GC_THREADS
#define GC_REDIRECT_TO_LOCAL
#include <gc/gc.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "types.h"
#include "env.h"
#include "int_funcs.h"
#include "functions.h"

#include "vm.h"
#include "vm.c"
#include "eval.h"
#include "rt.h"

#define ERROR_MSG_LENGTH 1024

#define OP_ARGS scm_vm * vm, scm_env * env, scm_value * arg1, unsigned char tail_position

static inline scm_value * list_reverse0(scm_value * list)
{
	scm_value * result = SCM_CONST_NIL;
	while (SCM_TYPE(list) != NIL) {
		result = SCM_CONS(SCM_CAR(list), result);
		list = SCM_CDR(list);
	}
	return result;
}

static inline scm_value * eval_symbol(scm_value * symbol, scm_env * env)
{
	scm_value * val;

	if (SCM_TYPE(symbol) != SYMBOL) return symbol;
	val = scm_env_get(env, SCM_SYMBOL(symbol));

	if (val != NULL) return val;
	
	/* else return error */
	char * errstr = (char *) GC_MALLOC(ERROR_MSG_LENGTH);
	sprintf(errstr, "Unbound symbol '%s'", SCM_SYMBOL(symbol));
	return scm_value_new_error(errstr);
}

static inline int load_args_into_env(scm_env * env, struct lambda * lambda, int argc, scm_value ** args)
{
	int i = 0;
	scm_value * arg_names = lambda->arg_list;
	while (1) {
		if (argc <= 0) break;
		if (SCM_TYPE(arg_names) == NIL) {
			break;
		} else if (SCM_TYPE(arg_names) == PAIR) {
			scm_env_add(env, SCM_SYMBOL(SCM_CAR(arg_names)), args[0]);
			arg_names = SCM_CDR(arg_names);
			args++;
			argc--;
			i++;
		} else if (SCM_TYPE(arg_names) == SYMBOL) {
			//scm_env_add(env, SCM_SYMBOL(arg_names), list_reverse0(list(argc - i, args)));
			//scm_env_add(env, SCM_SYMBOL(arg_names), list(argc - i, args));
			scm_env_add(env, SCM_SYMBOL(arg_names), list(argc, args));
			argc = 0;
			break;
		} else {
			printf("ay, caramba: %i\n", SCM_TYPE(arg_names));
		}
	}
	/* if argc < 1 -- too few arguments, argc > 1 too many arguments */
	return argc;
}

static inline scm_env * create_env(struct lambda * lambda, int argc, scm_value ** args, unsigned int local_symbols)
{
	return scm_env_new(lambda->parent, argc + local_symbols);
}

static inline scm_env * recycle_env(struct lambda * lambda, int argc, scm_value ** args, scm_env * env)
{
	scm_env_recycle(env);
	env->parent = lambda->parent;
	return env;
}

static inline void op_eval(OP_ARGS)
{
	if (SCM_TYPE(arg1) == PAIR) {
		vm_push_command_tail(vm, OP_INSPECT, env, SCM_CDR(arg1), tail_position);
		vm_push_command(vm, OP_EVAL, env, SCM_CAR(arg1));
	} else if (SCM_TYPE(arg1) == SYMBOL) {
		scm_value * val = eval_symbol(arg1, env);
		if ((SCM_TYPE(val)) != ERROR) vm_result_push(vm, val);
		else vm_error(vm, val);
	} else {
		scm_value * val = eval_symbol(arg1, env);
		if ((SCM_TYPE(val)) != ERROR) vm_result_push(vm, val);
		else vm_error(vm, val);
	}
}

/* transforms short define/define-macro form into lambda */
static inline scm_value * longer_define_form(scm_value * arg1)
{
	scm_value * args = SCM_CDAR(arg1);
	scm_value * code = SCM_CDR(arg1);
	scm_value * lmbd = SCM_CONS(scm_value_new_symbol("lambda"), SCM_CONS(args, code));
	return lmbd;
}

void push_args(scm_vm * vm, scm_env * env, scm_value * args)
{
	if (SCM_IS_NULL(args)) return;
	push_args(vm, env, SCM_CDR(args));
	vm_push_command_tail(vm, OP_EVAL, env, SCM_CAR(args), NOTAIL);
}

static inline void op_inspect(OP_ARGS)
{
	int length;
	enum scm_type type = SCM_TYPE(vm_result_top(vm));
	if ((type == FUNC) || (type == LAMBDA)) {
		vm_push_command_tail(vm, OP_RUN, env, scm_value_new_int(list_length(arg1)), tail_position);
		push_args(vm, env, arg1);
	}

	else if ((type == MACRO) || (type == PRECOMPILED_MACRO)) {
		vm_push_command(vm, OP_RETURN, env, NULL);
		vm_push_command(vm, OP_RUN, env, scm_value_new_int(list_length(arg1)));
		while (!SCM_IS_NULL(arg1)) {
			vm_result_push(vm, SCM_CAR(arg1));
			arg1 = SCM_CDR(arg1);
		}
	}

	else if (type == CONTINUATION) {
		if (list_length(arg1) != 1) {
			vm_error(vm, scm_value_new_error("Escape function requires 1 argument."));
			return;
		}

		vm_push_command(vm, OP_APPLY_CONTINUATION, NULL, vm_result_top(vm));
		vm_push_command(vm, OP_WAIT, NULL, NULL);
		vm_push_command(vm, OP_EVAL, env, SCM_CAR(arg1));

	} else if (type == SPEC_FORM) {
		scm_spec_type spec_form = SCM_SPEC_FORM(vm_result_pop(vm));
		switch (spec_form) {
			case SP_DEFINE:
				if (list_length(arg1) < 2) {
					vm_error(vm, scm_value_new_error("Syntax error in define"));
					return;
				}

				if (tail_position != HEAD)
					vm_push_command(vm, OP_DROP, NULL, NULL);
				vm_push_command(vm, OP_DEFINE, env, NULL);
				if (SCM_TYPE(SCM_CAR(arg1)) == SYMBOL) { 
					/* (define symbol (lambda ...))) */
					if (tail_position != HEAD)
						vm_push_command(vm, OP_WAIT, NULL, NULL);
					vm_push_command(vm, OP_EVAL, env, SCM_CADR(arg1));
					vm_result_push(vm, SCM_CAR(arg1));
				} else {
					/* (define (fun ...)) */
					if (SCM_TYPE(SCM_CAR(arg1)) != PAIR) {
						vm_error(vm, scm_value_new_error("Syntax error in define"));
						return;
					}

					scm_value * lmbd = longer_define_form(arg1);
					if (tail_position != HEAD)
						vm_push_command(vm, OP_WAIT, NULL, NULL);
					vm_push_command(vm, OP_EVAL, env, lmbd);
					vm_result_push(vm, SCM_CAAR(arg1));
				}
				break;
				
			case SP_MACRO:
				if (list_length(arg1) < 2) {
					vm_error(vm, scm_value_new_error("Syntax error in define-macro"));
					return;
				}
				vm_push_command(vm, OP_DROP, NULL, NULL);
				if (SCM_TYPE(SCM_CAR(arg1)) == SYMBOL) { 
					vm_push_command(vm, OP_DEFINE_MACRO, env, NULL);
					vm_push_command(vm, OP_WAIT, env, NULL); 
					vm_push_command(vm, OP_EVAL, env, SCM_CADR(arg1));
					vm_result_push(vm, SCM_CAR(arg1));
				} else {
					if (SCM_TYPE(SCM_CAR(arg1)) != PAIR) {
						vm_error(vm, scm_value_new_error("Syntax error in define"));
						return;
					}

					scm_value * lmbd = longer_define_form(arg1);
					vm_push_command(vm, OP_DEFINE_MACRO, env, NULL);
					vm_push_command(vm, OP_WAIT, env, NULL); 
					vm_push_command(vm, OP_EVAL, env, lmbd);
					vm_result_push(vm, SCM_CAAR(arg1));
				}
				break;

			case SP_SET:
				if (list_length(arg1) != 2) {
					vm_error(vm, scm_value_new_error("Spec. form. 'set!' requires 2 arguments"));
					return;
				}
				vm_push_command(vm, OP_DROP, env, SCM_CAR(arg1));
				vm_push_command(vm, OP_SET, env, NULL);
				//vm_push_command(vm, OP_WAIT, env, SCM_CAR(arg1));
				vm_push_command(vm, OP_WAIT, NULL, NULL);
				vm_push_command(vm, OP_EVAL, env, SCM_CADR(arg1));
				vm_result_push(vm, SCM_CAR(arg1));
				break;
				
			case SP_IF:
				length = list_length(arg1);
				if ((length != 2) && (length != 3)) {
					vm_error(vm, scm_value_new_error("Spec. form. 'if' requires 2 or 3 arguments"));
					return;
				}

				vm_push_command_tail(vm, OP_IF, env, NULL, tail_position);
				vm_push_command_tail(vm, OP_EVAL, env, SCM_CAR(arg1), NOTAIL);
				if (length == 2) vm_result_push(vm, SCM_CONST_VOID);
				else vm_result_push(vm, SCM_CADDR(arg1));
				vm_result_push(vm, SCM_CADR(arg1));
				break;
				
			case SP_LAMBDA: 
				{
					if (list_length(arg1) < 2) {
						vm_error(vm, scm_value_new_error("Syntax error in lambda expression"));
						return;
					}

					scm_value * ar = SCM_CAR(arg1);
					while (SCM_TYPE(ar) == PAIR) {
						if (SCM_TYPE(SCM_CAR(ar)) != SYMBOL) {
							vm_error(vm, scm_value_new_error("Argument of lambda expression have to be symbol"));
							return;
						}
						ar = SCM_CDR(ar);
					}
					if ((SCM_TYPE(ar) != NIL) && (SCM_TYPE(ar) != SYMBOL)) {
						vm_error(vm, scm_value_new_error("Argument of lambda expression have to be symbol"));
						return;
					}

					env->closure = 1; /* environment became a clousure */
					vm_result_push(vm, 
							scm_value_new_lambda(env, SCM_CAR(arg1), list_reverse0(SCM_CDR(arg1))));
					break;
				}
			case SP_QUOTE:
				if (list_length(arg1) != 1) {
					vm_error(vm, scm_value_new_error("Spec. form. 'quote' requires 1 argument"));
					return;
				}

				vm_result_push(vm, SCM_CAR(arg1));
				break;
			case SP_CALLCC:
				{
					if (list_length(arg1) != 1) {
						vm_error(vm, scm_value_new_error("Spec. form. 'call/cc' requires 1 argument"));
						return;
					}
					scm_value * val = scm_value_new_continuation(vm_clone(vm->runtime->main_vm));
					scm_value * lst = SCM_CONS(SCM_CAR(arg1), SCM_CONS(val, scm_value_new_nil()));
					vm_push_command(vm, OP_EVAL, env, lst);
				}
				break;
			case SP_MAP_EXP:
				if (list_length(arg1) != 2) {
					vm_error(vm, scm_value_new_error("Spec. form. 'map-exp' requires 2 arguments"));
					return;
				}
				vm_push_command(vm, OP_MAP_EXP, env, NULL);
				vm_push_command(vm, OP_EVAL, env, SCM_CADR(arg1));
				vm_result_push(vm, SCM_CAR(arg1));
				break;
			case SP_SETCAR:
				if (list_length(arg1) != 2) {
					vm_error(vm, scm_value_new_error("Spec. form. 'set-car!' requires 2 arguments"));
					return;
				}
				vm_push_command(vm, OP_DROP, env, SCM_CAR(arg1));
				vm_push_command(vm, OP_SETCAR, env, NULL);
				vm_push_command(vm, OP_WAIT, NULL, NULL);
				//vm_push_command(vm, OP_WAIT, env, SCM_CAR(arg1));
				vm_push_command(vm, OP_EVAL, env, SCM_CAR(arg1));
				vm_push_command(vm, OP_EVAL, env, SCM_CADR(arg1));
				break;
			case SP_SETCDR:
				if (list_length(arg1) != 2) {
					vm_error(vm, scm_value_new_error("Spec. form. 'set-car!' requires 2 arguments"));
					return;
				}
				vm_push_command(vm, OP_DROP, env, SCM_CAR(arg1));
				vm_push_command(vm, OP_SETCDR, env, NULL);
				vm_push_command(vm, OP_WAIT, NULL, NULL);
				//vm_push_command(vm, OP_WAIT, env, SCM_CAR(arg1));
				vm_push_command(vm, OP_EVAL, env, SCM_CAR(arg1));
				vm_push_command(vm, OP_EVAL, env, SCM_CADR(arg1));
				break;
			case SP_VECTOR_SET:
				if (list_length(arg1) != 3) {
					vm_error(vm, scm_value_new_error("Spec. form. 'vector-set!' requires 3 arguments"));
					return;
				}
				vm_push_command(vm, OP_DROP, env, SCM_CAR(arg1));
				vm_push_command(vm, OP_VECTOR_SET, env, NULL);
				vm_push_command(vm, OP_WAIT, NULL, NULL);
			//	vm_push_command(vm, OP_WAIT, env, SCM_CAR(arg1));
				vm_push_command(vm, OP_EVAL, env, SCM_CAR(arg1));
				vm_push_command(vm, OP_EVAL, env, SCM_CADR(arg1));
				vm_push_command(vm, OP_EVAL, env, SCM_CADDR(arg1));
				break;
			case SP_DISPLAY:
				if (list_length(arg1) != 1) {
					vm_error(vm, scm_value_new_error("Spec. form. 'display' requires 1 argument"));
					return;
				}
				vm_push_command(vm, OP_DISPLAY, NULL, NULL);
				vm_push_command(vm, OP_WAIT, NULL, NULL);
				vm_push_command(vm, OP_EVAL, env, SCM_CAR(arg1));
				break;
			case SP_NEWLINE:
				if (list_length(arg1) != 0) {
					vm_error(vm, scm_value_new_error("Spec. form. 'newline' requires no argument"));
					return;
				}
				vm_push_command(vm, OP_NEWLINE, NULL, NULL);
				vm_push_command(vm, OP_WAIT, NULL, NULL);
				break;


		}
	} else {
		vm_error(vm, scm_value_new_error("Wrong type to apply"));
	}
}

static inline void load_commads(scm_vm * vm, scm_env * new_env, scm_value * cmds)
{
	int tail = 1;

	while (SCM_TYPE(cmds) != NIL) {
		if (tail) {
			vm_push_command_tail(vm, OP_EVAL, new_env, SCM_CAR(cmds), TAIL);
			tail = 0;
		} else vm_push_command_tail(vm, OP_EVAL, new_env, SCM_CAR(cmds), NOTAIL);
		cmds = SCM_CDR(cmds);
		if (!SCM_IS_NULL(cmds))
			vm_push_command(vm, OP_DISCARD, NULL, NULL);
	}
}
static void op_run(OP_ARGS)
{
	int i;
	int argc = SCM_INT(arg1);
	scm_value * args[argc];
	scm_value * func_val;

	for (i = argc - 1; i >= 0; i--)
		args[i] = vm_result_pop(vm);

	func_val = vm_result_pop(vm);

	if ((SCM_TYPE(func_val) == FUNC) || (SCM_TYPE(func_val) == PRECOMPILED_MACRO)) {
		scm_func func = SCM_FUNC(func_val);
		scm_value * result = func(argc, args);
		if (SCM_TYPE(result) == ERROR) vm_error(vm, result);
		else vm_result_push(vm, result);
	} else if ((SCM_TYPE(func_val) == LAMBDA) || (SCM_TYPE(func_val) == MACRO) 
			|| (SCM_TYPE(func_val) == CONTINUATION)) {

		struct lambda * lmbd = SCM_LAMBDA(func_val);
		scm_env * new_env = ((tail_position && (!env->closure) && (SCM_TYPE(func_val) != LAMBDA)) 
				? recycle_env(lmbd, argc, args, env)
				//? create_env(lmbd, argc, args, lmbd->local_symbols)
				: create_env(lmbd, argc, args, lmbd->local_symbols));

		int state = load_args_into_env(new_env, lmbd, argc, args);
		if (state != 0) {
			vm_error(vm, scm_value_new_error("Wrong count of arguments"));
			return;
		}

		if (SCM_IS_NULL(lmbd->cmd_list)) {
			vm_error(vm, scm_value_new_error("Lambda expression missing body"));
			return;
		}

		scm_value * cmds = lmbd->cmd_list;
		if (SCM_IS_NULL(lmbd->define_list)) load_commads(vm, new_env, cmds);
		else {
			vm_push_command_tail(vm, OP_BLOCK, new_env, cmds, HEAD);
			vm_push_command(vm, OP_DISCARD, NULL, NULL);
			/*
			load_commads(vm, new_env, cmds);
			vm_push_command(vm, OP_DISCARD, NULL, NULL);
			*/	
		}
		
		cmds = lmbd->define_list;

		while (!SCM_IS_NULL(cmds)) {
			vm_push_command_tail(vm, OP_EVAL, new_env, SCM_CAR(cmds), HEAD);
			cmds = SCM_CDR(cmds);
			if (!SCM_IS_NULL(cmds)) 
				vm_push_command(vm, OP_DISCARD, NULL, NULL); // potrebuje aby fce mela telo???
		}
	} else {
		printf("ay, caramba! i hate %i as type.\n", SCM_TYPE(func_val));
	}
}

static inline void op_define(OP_ARGS)
{
	scm_value * val = vm_result_pop(vm);
	scm_value * symbol = vm_result_pop(vm);
	if (SCM_TYPE(symbol) != SYMBOL) {
		vm_error(vm, scm_value_new_error("First argument of define must be symbol"));
		return;
	}

	scm_env_add(env, SCM_SYMBOL(symbol), val);
	vm_result_push(vm, val);
}

static inline void op_set(OP_ARGS)
{
	scm_value * val = vm_result_pop(vm);
	scm_value * symbol = vm_result_pop(vm);

	if (SCM_TYPE(symbol) != SYMBOL) {
		vm_error(vm, scm_value_new_error("First argument of set! must be symbol"));
		return;
	}

	scm_env_set(env, SCM_SYMBOL(symbol), val);
	vm_result_push(vm, val);
}

static inline void op_setcar(OP_ARGS)
{
	scm_value * pair = vm_result_pop(vm);
	scm_value * val = vm_result_pop(vm);

	if (SCM_TYPE(pair) != PAIR) {
		vm_error(vm, scm_value_new_error("First argument of set-car! must be pair"));
		return;
	}

	SCM_PAIR(pair).ar = val;
	vm_result_push(vm, pair);
}

static inline void op_setcdr(OP_ARGS)
{
	scm_value * pair = vm_result_pop(vm);
	scm_value * val = vm_result_pop(vm);

	if (SCM_TYPE(pair) != PAIR) {
		vm_error(vm, scm_value_new_error("First argument of set-cdr! must be pair"));
		return;
	}

	SCM_PAIR(pair).dr = val;
	vm_result_push(vm, pair);
}

static inline void op_vector_set(OP_ARGS)
{
	scm_value * vector = vm_result_pop(vm);
	scm_value * k = vm_result_pop(vm);
	scm_value * val = vm_result_pop(vm);

	if (SCM_TYPE(vector) != VECTOR) {
		vm_error(vm, scm_value_new_error("First argument of vector-set! must be pair"));
		return;
	}

	if (SCM_TYPE(vector) != VECTOR) {
		vm_error(vm, scm_value_new_error("Second argument of vector-set! must be integer."));
		return;
	}

	int k1 = SCM_INT(k);
	if ((k1 < 0) || (k1 >= SCM_VECTOR(vector)->size)) {
		vm_error(vm, scm_value_new_error("Index out of range in vector-set!"));
		return;
	}

	SCM_VECTOR(vector)->items[k1] = val;
	vm_result_push(vm, vector);
}

static scm_value * build_map_exp(scm_value * func, scm_value * val)
{
	if (SCM_IS_NULL(val)) return val;
	scm_value * qouted = scm_value_new_pair(scm_value_new_symbol("quote"), scm_value_new_pair(SCM_CAR(val), scm_value_new_nil()));
	return scm_value_new_pair(
			scm_value_new_pair(func, scm_value_new_pair(qouted,
					scm_value_new_nil())),
				build_map_exp(func, SCM_CDR(val)));
}

static inline void op_map_exp(OP_ARGS)
{
	scm_value * val = vm_result_pop(vm);
	scm_value * function = vm_result_pop(vm);

	scm_value * result = build_map_exp(function, val);
	result = scm_value_new_pair(scm_value_new_symbol("list"), result);

	vm_push_command_tail(vm, OP_EVAL, env, result, tail_position);
}

static inline void op_define_macro(OP_ARGS)
{			
	scm_value * val = vm_result_pop(vm);
	scm_value * symbol = vm_result_pop(vm);

	if (SCM_TYPE(val) != LAMBDA) {
		vm_error(vm, scm_value_new_error("Second argument of 'define-macro' must be user defined function."));
		return;
	}
	val = scm_value_new_lambda_clone(val);
	val->type = MACRO;
	scm_env_add(env, SCM_SYMBOL(symbol), val);
	vm_result_push(vm, val);
}

static inline void op_if(OP_ARGS)
{
	scm_value * result = scm_value_new_void();
	scm_value * val = vm_result_pop(vm);
	scm_value * result1 = vm_result_pop(vm);
	scm_value * result2 = vm_result_pop(vm);

	if ((SCM_TYPE(val) != BOOL) && (SCM_TYPE(val) != FLOAT) && (SCM_TYPE(val) != INT) && (SCM_TYPE(val) != NIL)) {
		vm_error(vm, scm_value_new_error("Expecting boolean or number as 'if' first argument"));
		return;
	}
	if (SCM_TYPE(val) == BOOL) result = (SCM_BOOL(val) ? result1 : result2);
	if (SCM_TYPE(val) == INT) result = (SCM_INT(val) ? result1 : result2);
	if (SCM_TYPE(val) == FLOAT) result = (SCM_FLOAT(val) ? result1 : result2);
	if (SCM_TYPE(val) == NIL) result = result1;

	//scm_value * result = (SCM_BOOL(val) ? result1 : result2);
	vm_push_command_tail(vm, OP_EVAL, env, result, tail_position);
}

static inline void op_discard(OP_ARGS)
{
	vm_result_pop(vm);
}

static inline void op_return(OP_ARGS)
{
	scm_value * val = vm_result_pop(vm);
	vm_push_command(vm, OP_EVAL, env, val);
}

static inline void op_block(OP_ARGS)
{
	load_commads(vm, env, arg1);
}

static inline void op_feval(scm_vm * vm, command * cmd)
{
	scm_vm * master = vm;
	scm_vm * slave = cmd->vm;

	vm_lock(slave);
	vm_join(master, slave);
	vm_unlock(slave);
}

static inline void op_drop(scm_vm * vm, command * cmd)
{
	vm_drop_fevals(vm);
}

static inline void op_apply_continuation(scm_vm * vm, scm_value * arg1)
{
	scm_vm * cont = arg1->value.continuation;
	scm_value * esc = vm_result_pop(vm);
	stack_item item;
	while (!stack_empty(vm->exct)) stack_pop(vm->exct);
	while (!stack_empty(vm->rslt)) stack_pop(vm->rslt);

	item = stack_internal_get_bottom(cont->rslt);
	while (!stack_internal_get_done(cont->rslt, item)) {
		vm_result_push(vm, stack_internal_get_value(cont->rslt, item));
		item = stack_internal_get_prev(cont->rslt, item);
	}

	item = stack_internal_get_bottom(cont->exct);
	while (!stack_internal_get_done(cont->exct, item)) {
		command * cmd = (command *)stack_internal_get_value(cont->exct, item);
		vm_push_command_tail(vm, cmd->op, cmd->env, cmd->arg1, cmd->tail_position);
		item = stack_internal_get_prev(cont->exct, item);
	}

	vm_result_push(vm, esc);
}

static inline void op_display(OP_ARGS)
{
	scm_value * val = vm_result_pop(vm);
	display_value(val);
	vm_result_push(vm, scm_value_new_void());
}

static inline void op_newline(OP_ARGS)
{
	printf("\n");
	vm_result_push(vm, scm_value_new_void());
}


void * eval_vm(void * arg_vm)
{
	command * top;
	scm_value * arg1;
	scm_env * arg_env;
	scm_vm * vm = (scm_vm *)arg_vm;
	scm_rt * rt = vm_get_runtime(vm);
	unsigned int limiter = 0;
 	unsigned short reopen = 0;
	unsigned char tail_position;
	int done = 0;

	vm_lock(vm);
	while (!vm_done(vm)) {
		top = vm_pop_command(vm);
		arg1 = top->arg1;
		arg_env = top->env;
		tail_position = top->tail_position; 

		switch (top->op) {
			case OP_EVAL: op_eval(vm, arg_env, arg1, tail_position); break;
			case OP_INSPECT: op_inspect(vm, arg_env, arg1, tail_position); break;
			case OP_RUN: op_run(vm, arg_env, arg1, tail_position); break;
			case OP_DEFINE: op_define(vm, arg_env, arg1, tail_position); break;
			case OP_SET: op_set(vm, arg_env, arg1, tail_position); break;
			case OP_DEFINE_MACRO: op_define_macro(vm, arg_env, arg1, tail_position); break;
			case OP_IF: op_if(vm, arg_env, arg1, tail_position); break;
			case OP_DISCARD: op_discard(vm, arg_env, arg1, tail_position); break;
			case OP_RETURN: op_return(vm, arg_env, arg1, tail_position); break;
			case OP_APPLY_CONTINUATION: op_apply_continuation(vm, arg1); break;
			case OP_MAP_EXP: op_map_exp(vm, arg_env, arg1, tail_position); break;
			case OP_SETCAR: op_setcar(vm, arg_env, arg1, tail_position); break;
			case OP_SETCDR: op_setcdr(vm, arg_env, arg1, tail_position); break;
			case OP_VECTOR_SET: op_vector_set(vm, arg_env, arg1, tail_position); break;
			case OP_FEVAL:
				op_feval(vm, top); 
				limiter = rt->scheduler_limit + 1;
				break;
			case OP_WAIT: 
				if (vm->parent == NULL) done = 0;
				else done = 1;
				break;
			case OP_DROP: op_drop(vm, top); break;
			case OP_BLOCK: op_block(vm, arg_env, arg1, tail_position); break; 
			case OP_DISPLAY: op_display(vm, arg_env, arg1, tail_position); break; 
			case OP_NEWLINE: op_newline(vm, arg_env, arg1, tail_position); break; 
		}
		if (!done) vm_command_free(vm, top);

		if (limiter < rt->scheduler_limit) limiter++;
		else {
			rt_schedule_fevals(rt, vm);
			limiter = 0;
		}
		if (reopen < rt->reopen_limit) reopen++;
		else {
			vm_unlock(vm);
			vm_lock(vm);
			reopen = 0;
		}
		if (done) break;
	}
	rt_lock(rt);
	rt->sched_count--;
	rt_log(rt);
	rt_unlock(rt);
	vm_unlock(vm);
	return NULL;
}
