#define GC_REDIRECT_TO_LOCAL
#include <gc/gc.h>
#include <stdlib.h>
#include "types.h"
#include "int_funcs.h"
#include <string.h>

#define ALLOC_VAL		((scm_value *) GC_MALLOC(sizeof(scm_value)))
#define ALLOC_ATOMIC_VAL	((scm_value *) GC_MALLOC_ATOMIC(sizeof(scm_value)))

scm_value * scm_const_nil = NULL;
scm_value * scm_const_true = NULL;
scm_value * scm_const_false = NULL;
scm_value * scm_const_void = NULL;

void scm_init_constants()
{
	scm_const_nil = scm_value_new_nil();
	scm_const_false = scm_value_new_bool(0);
	scm_const_true = scm_value_new_bool(1);
	scm_const_void = scm_value_new_void();
}

#ifndef INTEGER_POINTER_HACK
scm_value * scm_value_new_int(int val) 
{
	scm_value * result = ALLOC_ATOMIC_VAL;
	result->type = INT;
	result->value.integer = val;
	return result;
}
#else
/*
defined as macro in types.h
*/
#endif

scm_value * scm_value_new_symbol(char * symbol) 
{
	scm_value * result = ALLOC_VAL;
	result->type = SYMBOL;
	result->value.symbol = symbol;
	return result;
}

scm_value * scm_value_new_float(double val) 
{
	scm_value * result = ALLOC_ATOMIC_VAL;
	result->type = FLOAT;
	result->value.flt = val;
	return result;
}

scm_value * scm_value_new_func(scm_func func) 
{
	scm_value * result = ALLOC_ATOMIC_VAL;
	result->type = FUNC;
	result->value.func = func;
	return result;
}

scm_value * scm_value_new_spec_form(scm_spec_type type)
{
	scm_value * result = ALLOC_ATOMIC_VAL;
	result->type = SPEC_FORM;
	result->value.spec_form = type;
	return result;
}

scm_value * scm_value_new_precompiled_macro(scm_func type)
{
	scm_value * result = ALLOC_ATOMIC_VAL;
	result->type = PRECOMPILED_MACRO;
	result->value.func = type;
	return result;
}

scm_value * scm_value_new_pair(scm_value * ar, scm_value * dr) 
{
	scm_value * result = ALLOC_VAL;
	result->type = PAIR;
	result->value.pair.ar = ar;
	result->value.pair.dr = dr;
	return result;
}

scm_value * scm_value_new_bool(int value) 
{
	scm_value * result = ALLOC_ATOMIC_VAL;
	result->type = BOOL;
	result->value.bool = value;
	return result;
}

scm_value * scm_value_new_nil()
{
	scm_value * result = ALLOC_ATOMIC_VAL;
	result->type = NIL;
	return result;
}

scm_value * scm_value_new_lambda(scm_env * env, scm_value * arg_list, scm_value * cmd_list)
{
	scm_value * result = ALLOC_VAL;
	struct lambda * fun = (struct lambda *) GC_MALLOC(sizeof(struct lambda));
	result->type = LAMBDA;
	fun->parent = env;
	fun->arg_list = arg_list;
	fun->cmd_list = cmd_list;
	fun->define_list = scm_value_new_nil();

	scm_value * reversed = list_reverse(cmd_list);
	while ((SCM_TYPE(reversed) == PAIR) && (SCM_TYPE(SCM_CAR(reversed)) == PAIR) && 
		(!strcmp(SCM_SYMBOL(SCM_CAAR(reversed)), "define"))) {
		fun->define_list = scm_value_new_pair(SCM_CAR(reversed), fun->define_list);
		reversed = SCM_CDR(reversed);
		fun->local_symbols++;
	}
	fun->cmd_list = list_reverse(reversed);
	result->value.lambda = fun; 
	return result;
}

scm_value * scm_value_new_lambda_clone(scm_value * val)
{
	scm_value * result = ALLOC_VAL;
	struct lambda * fun = (struct lambda *) GC_MALLOC(sizeof(struct lambda));
	memcpy(fun, val->value.lambda, sizeof(struct lambda));
	result->type = LAMBDA;
	result->value.lambda = fun;
	return result;
}

scm_value * scm_value_new_void()
{
	scm_value * result = ALLOC_ATOMIC_VAL;
	result->type = VOID;
	return result;
}

scm_value * scm_value_new_error(char * desc)
{
	scm_value * result = ALLOC_VAL;
	result->type = ERROR;
	result->value.err_desc = desc;
	return result;
}

scm_value * scm_value_new_continuation(struct scm_vm * continuation)
{
	scm_value * result = ALLOC_VAL;
	result->type = CONTINUATION;
	result->value.continuation = continuation;
	return result;
}

static inline scm_value * empty_vector(int length)
{
	scm_value * result = ALLOC_VAL;
	scm_vector * vector = (scm_vector *)GC_MALLOC(sizeof(scm_vector));
	vector->size = length;
	vector->items = GC_MALLOC(sizeof(scm_value *) * length);
	result->type = VECTOR;
	result->value.vector = vector;
	return result;
}

scm_value * scm_value_new_vector_from_list(scm_value * items)
{
	int length = list_length(items);
	scm_value * result = empty_vector(length);
	int i = 0;

	while (SCM_TYPE(items) == PAIR) {
		SCM_VECTOR(result)->items[i] = SCM_CAR(items);
		i++;
		items = SCM_CDR(items);
	}
	return result;
}

scm_value * scm_value_new_vector(int count, scm_value * val)
{
	int i;
	scm_value * result = empty_vector(count);
	
	for (i = 0; i < count; i++)
		SCM_VECTOR(result)->items[i] = val;
	return result;
}
