#ifndef VM_C
#define VM_C
#define GC_REDIRECT_TO_LOCAL
#include <gc/gc.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "mutex.h"
#include "types.h"
#include "env.h"
#include "vm.h"
#include "rt.h"
#include "stack.c"

static inline scm_vm * vm_new(scm_rt * runtime)
{
	scm_vm * result = (scm_vm *)GC_MALLOC(sizeof(scm_vm));
	result->runtime = runtime;
	result->exct = stack_new();
	result->rslt = stack_new();
	result->parent = NULL;
	result->command_pool = NULL;
	result->vm_id = (unsigned long *)GC_MALLOC(sizeof(unsigned long));

	rt_lock(runtime);
	runtime->last_id++;
	*(result->vm_id) = runtime->last_id;
	rt_unlock(runtime);
	mutex_init(result->lock);
	return result;
}

static inline scm_vm * vm_new_init(scm_rt * runtime, command * initial_command)
{
	scm_vm * result = vm_new(runtime);
	vm_push_command(result, initial_command->op, initial_command->env, initial_command->arg1);
	return result;
}

/** vraci 0 pokud nejsou dalsi prikazy ke zpracovani */
static inline int vm_done(scm_vm * vm)
{
	return (stack_empty(vm->exct));
}

static inline void vm_push_command(scm_vm * vm, enum command_op op, scm_env * env, scm_value * arg)
{
	stack_push(vm->exct, vm_command_new(vm, op, env, arg, NOTAIL));
}

static inline void vm_push_command_tail(scm_vm * vm, enum command_op op, scm_env * env, scm_value * arg, unsigned char tail_position)
{
	stack_push(vm->exct, vm_command_new(vm, op, env, arg, tail_position));
}

static inline command * vm_pop_command(scm_vm * vm)
{
	return stack_pop(vm->exct);
}

static inline command * vm_command_new(scm_vm * vm, enum command_op op, scm_env * env, scm_value * arg, unsigned char tail_position)
{
	command * cmd;
	if (vm->command_pool == NULL) cmd = (command *)GC_MALLOC(sizeof(command));
	else {
		cmd = vm->command_pool;
		vm->command_pool = vm->command_pool->next_pooled;
	}

	cmd->env = env;
	cmd->op = op;
	cmd->arg1 = arg;
	cmd->tail_position = tail_position;	
	return cmd;
}

static inline command * vm_command_new0(enum command_op op, scm_env * env, scm_value * arg)
{
	command * cmd = (command *)GC_MALLOC(sizeof(command));

	cmd->env = env;
	cmd->op = op;
	cmd->arg1 = arg;
	
	return cmd;
}

static inline void vm_command_free(scm_vm * vm, command * cmd)
{
	cmd->next_pooled = vm->command_pool;
	vm->command_pool = cmd;
}

/** vraci prvek na pozici v navratovem zasobniku */
static inline scm_value * vm_result_top(scm_vm * vm)
{
	return stack_top(vm->rslt);
}

static inline scm_value * vm_result_pop(scm_vm * vm)
{
	return stack_pop(vm->rslt);
}

static inline void vm_result_push(scm_vm * vm, scm_value * value)
{
	stack_push(vm->rslt, value);
}

/* Appends appropriate stacks from slave vm into master vm, assumes taht all vms are locked */
void vm_join(scm_vm * master, scm_vm * slave)
{
	stack_append(master->exct, slave->exct);
	stack_append(master->rslt, slave->rslt);
	*(slave->vm_id) = *(master->vm_id);
}

static inline void vm_lock(scm_vm * vm) 
{
	mutex_lock(vm->lock);
}

static inline void vm_unlock(scm_vm * vm)
{
	mutex_unlock(vm->lock);
}

static inline scm_rt * vm_get_runtime(scm_vm * vm)
{
	return vm->runtime;
}

static void drop_thread(scm_vm * vm)
{
	vm_lock(vm);
	while (!stack_empty(vm->exct)) {
		command * cmd = vm_pop_command(vm);
		if (cmd->op == OP_FEVAL) {
			cmd->op = OP_EVAL;
			drop_thread(cmd->vm);
		}
	}
	vm_unlock(vm);
}

static inline void drop_all_threads(scm_vm * vm)
{
	stack_item exct = stack_internal_get_top(vm->exct);
	while (!stack_internal_get_done(vm->exct, exct)) {
		command * cmd = (command *)stack_internal_get_value(vm->exct, exct);
		if (cmd->op == OP_FEVAL) {
			fprintf(stderr, "DROP\n");
			cmd->op = OP_EVAL;
			drop_thread(cmd->vm);
		} 
		exct = stack_internal_get_next(vm->exct, exct);
	}
}

/* assumes locked vm */
void vm_drop_fevals(scm_vm * vm)
{
	vm->runtime->last_drop++;
	drop_all_threads(vm);
}

static inline void vm_error(scm_vm * vm, scm_value * error)
{
	drop_all_threads(vm);
	while (!stack_empty(vm->exct))
		stack_pop(vm->exct);
	while (!stack_empty(vm->rslt))
		stack_pop(vm->rslt);
	stack_push(vm->rslt, error);

}

static inline scm_vm * vm_clone(scm_vm * vm)
{
	scm_vm * clon = vm_new(vm->runtime);

	stack_item item = stack_internal_get_bottom(vm->rslt);
	while (!stack_internal_get_done(vm->rslt, item)) {
		vm_result_push(clon, stack_internal_get_value(vm->rslt, item));
		item = stack_internal_get_prev(vm->rslt, item);
	}

	item = stack_internal_get_bottom(vm->exct);
	while (!stack_internal_get_done(vm->exct, item)) {
		command * cmd = (command *)stack_internal_get_value(vm->exct, item);
		if (cmd->op != OP_FEVAL)
			vm_push_command_tail(clon, cmd->op, cmd->env, cmd->arg1, cmd->tail_position);
		else
			vm_push_command_tail(clon, OP_EVAL, cmd->env, cmd->arg1, cmd->tail_position);
		item = stack_internal_get_prev(vm->exct, item);
	}

	return clon;
}
#endif
